@@ -31,9 +31,11 @@
 #include <unistd.h>
 #include <errno.h>
 
-#include <bluetooth/bluetooth.h>
 #include <glib.h>
 
+#include "lib/bluetooth.h"
+#include "lib/sdp.h"
+
 #include "src/plugin.h"
 #include "src/adapter.h"
 #include "src/device.h"
@@ -108,7 +110,7 @@ static ssize_t autopair_pincb(struct btd_adapter *adapter,
 			if (attempt >= 4)
 				return 0;
 
-			snprintf(pinstr, sizeof(pinstr), "%06d",
+			snprintf(pinstr, sizeof(pinstr), "%06u",
 						rand() % 1000000);
 			*display = true;
 			memcpy(pinbuf, pinstr, 6);
